-- This file is part of the reachability analysis tool LaRA
-- (Lazy Reachability Analyzer)
--
-- Loïg Jezequel & Didier Lime
--
-- University of Nantes, École Centrale de Nantes, All rights reserved, 2016

module Lara.Data (
    PNode(..),
    Automaton(..),
    Network(..),
    Config(..),
    EType(..),
    AGraph,
    Objective,
    showAuto,
    updateLabel,
    updateELabel,
    initial,
    mergeNodes,
    gLab,
    edgeUnion,
    edgeUnionL,
    lookUpUnion,
    independent
    ) where

-- import GHC.Stack (errorWithStackTrace)
import Data.List
import Data.Maybe
import qualified Data.Map.Strict as M
import qualified Data.Set as S
import qualified Data.Sequence as Q

import qualified Data.Graph.Inductive.Graph as G
import qualified Data.Graph.Inductive.PatriciaTree as GP

import Control.Applicative

import Debug.Trace

import Lara.Time

--------------------------------------------------------------------------------
--                               Data-structures
--------------------------------------------------------------------------------
type Config = [Int]

-- edge annotations on the automata
data EType = EType { elab   :: Int,           -- the label
                     timing :: Maybe TimeInfo -- the timing info with a special 
                                              -- case for untimed edges
                   } deriving (Eq ,Ord, Show)

-- a PAutomaton Node
data PNode = PNode { config   :: (Config, Maybe DBM), -- current symbolic state
                     goal     :: Bool,                -- wether this is a goal state
                     autoReqs :: S.Set Int,           -- automata synced to reach this node
                     clkReqs  :: S.Set Int,           -- clocks read to reach this node
                     label    :: String,              -- a label derived from the individual states
                     predNs   :: [G.Node],            -- predecessor nodes (trace) in the spanning reachability tree
                     inv      :: Maybe TimeInfo,      -- possibly an invariant
                     actClks  :: M.Map Int Int        -- give for each clock its max const before reset
                   } deriving (Eq, Ord)

-- Underlying graph of an automaton
type AGraph = GP.Gr PNode EType

-- An automaton representing the (partial) product of automata or just one automaton
data Automaton = Automaton { graph        :: AGraph,                           -- the graph itself
                             nodeMap      :: M.Map Config  (G.Node, Q.Seq DBM),    -- which node and list of DBMs correspond to each discrete config
                             autoIDs      :: S.Set Int,                        -- the automata involved in this (product) automaton
                             goals        :: S.Set G.Node,                     -- the set of nodes which are goals
                             name         :: String,                           -- a name for this automaton
                             laReqs       :: M.Map Int (S.Set Int),            -- which automata also use the labels used in this automaton
                             hasObj       :: Bool,                             -- whether this automaton contains a part of the objective
                             timed        :: Bool,                             -- whether this is a timed automaton
                             aclocks      :: S.Set Int,                        -- the set of clocks used in this automaton
                             clockMap     :: M.Map Int Int,                    -- how clocks are mapped in DBMs for this automaton
                             confForcers  :: M.Map Int (S.Set Int),            -- which automata can force the labels used in this automaton
                             resetForcers :: M.Map Int (S.Set Int),            -- which automata can force the reset of clocks read in this automaton
                             resetters    :: M.Map Int (S.Set Int),            -- which automata can reset the clocks read in this automaton
                             kByClock     :: M.Map Int Int                     -- the maximum constants for clocks used in this automaton
                           } deriving (Eq, Show)


-- An objective given as a list of pairs (Automaton, State of the automaton)
type Objective = [(Int, Int)]

data Network = Network { automata     :: [Automaton],           -- a list of all the automata in the system
                         labels       :: [String],              -- the labels of all actions in the system
                         clocks       :: [String],              -- the names of all clocks in the system
                         objective    :: Objective,             -- the objective to fulfill
                         lnReqs       :: M.Map Int (S.Set Int), -- which automata use each label
                         labUrgency   :: M.Map Int (S.Set Int), -- which automata can force each label
                         clockUrgency :: M.Map Int (S.Set Int), -- which automata can force the reset of each clock
                         clkResets    :: M.Map Int (S.Set Int), -- which automata can reset each clocks
                         nbautos      :: Int                    -- the number of automata in the system (for performance reasons)
                       } deriving Show
   

--------------------------------------------------------------------------------
--                                  Display
--------------------------------------------------------------------------------
showEdge :: AGraph -> [String] -> (G.Node, EType) -> String
showEdge g ls (n',i) = "\n    -- " 
                        ++ ls !! elab i 
                        ++ (if isJust (timing i) then " | " ++ show (fromJust $ timing i) 
                                                 else "")
                        ++ " -> " 
                        ++ show n' ++ ": " ++ show (gLab g n')

showNode :: AGraph -> [String] -> G.Node -> String
showNode g ls n = show n ++": "++ show (gLab g n) ++ concatMap (showEdge g ls) (G.lsuc g n) ++ "\n"

showAuto :: Network -> Automaton -> String
showAuto net aut = let g = graph aut
                       ns = G.nodes g
                    in concatMap (showNode g (labels net)) ns

instance Show PNode where
    show pn = label pn ++ (if goal pn then "!" else "") ++ " " ++ show (S.toList $ autoReqs pn)
                                                        ++ " + " ++ show (S.toList $ clkReqs pn)
                                                        -- ++ "\n" ++ show (snd $ config pn)


--------------------------------------------------------------------------------
--                                  Misc
--------------------------------------------------------------------------------

gLab :: AGraph -> G.Node -> PNode
gLab g n = --fromJust.(G.lab g)
           fromMaybe (error $ "cannot get label for unknown node " ++ show n) $ G.lab g n


-- update the label of a node in a graph
updateLabel :: G.Node -> PNode -> AGraph -> AGraph
updateLabel n l' g = case G.match n g of
      (Just (p, _, l, s), cg) -> (p, n, l', s) G.& cg
      (Nothing, _)            -> error $ "Unknown node" ++ show n

-- update the label of an edge in a graph
updateELabel :: G.LEdge EType -> EType -> AGraph -> AGraph
updateELabel (n,n',i) i' g = case G.match n g of
      (Just (p, _, l, s), cg) -> (p, n, l, map (\e@(i'', n'') -> if n'' == n' && i'' == i' 
                                                                 then (i', n') 
                                                                 else e) s) G.& cg
      (Nothing, _)            -> error $ "Unknown node" ++ show n

-- Union of two ETypes
edgeUnion :: EType -> EType -> EType
edgeUnion (EType i mts) (EType i' mts') =
     if i == i' then EType i $ mtiUnion mts mts'
                else error "Cannot take the union of two edges with different labels"

-- Union of two ETypes
edgeUnionL :: [EType] -> EType
edgeUnionL [] = error "Cannot take the union of 0 edges"
edgeUnionL xs = foldl1' edgeUnion xs


-- initial node of a Automaton
initial :: Automaton -> Config
initial = map (const 0) . S.toList . autoIDs

-- merge helper
mergeN' :: [(Int,Int)] -> [(Int,Int)] -> Maybe [(Int,Int)]
mergeN' [] ys = Just ys
mergeN' xs [] = Just xs
mergeN' xs@((i1,s1):xs') ys@((i2,s2):ys')
  | i1 == i2  = if s1 == s2 then liftA2 (:) (Just (i1,s1)) (mergeN' xs' ys')
                            else Nothing
  | i1 < i2   = liftA2 (:) (Just (i1,s1)) (mergeN' xs' ys)
  | otherwise = liftA2 (:) (Just (i2,s2)) (mergeN' xs ys')

-- Merge Config's from two Automaton's
mergeNodes :: Config -> [Int] -> Config -> [Int] -> Maybe [Int]
mergeNodes c1 as1 c2 as2 = 
    let 
        x1 = zip as1 c1
        x2 = zip as2 c2 
    in
        map snd <$> mergeN' x1 x2

-- find the union of all sets r[l] such that l in i
lookUpUnion :: S.Set Int -> M.Map Int (S.Set Int) -> S.Set Int
lookUpUnion i r = S.foldl' (\s l -> S.union s (M.findWithDefault S.empty l r)) S.empty i

-- is action i from aut1 independent from aut2?
independent :: Automaton -> EType -> Automaton -> Bool
independent aut1 i aut2 =
    let sas2 = autoIDs aut2
     in S.null $ S.intersection sas2 $ laReqs aut1 M.! elab i
        

