-- This file is part of the reachability analysis tool LaRA
-- (Lazy Reachability Analyzer)
--
-- Loïg Jezequel & Didier Lime
--
-- University of Nantes, École Centrale de Nantes, All rights reserved, 2016

module Lara.PReach 
    (preach,
     computeTrace,
     preachFull
    ) where

import Data.Ord
import Data.Function
import Data.List
import Data.Maybe (fromJust, fromMaybe)
import qualified Data.Graph.Inductive.Graph as G
import qualified Data.Graph.Inductive.PatriciaTree as GP
import qualified Data.Set as S
import qualified Data.Map.Strict as M
import qualified Data.Sequence as Q
import qualified Data.Foldable as F (concat)

import Control.Applicative

import System.Environment
import Debug.Trace

import Lara.Data
import Lara.Time
import Lara.Product

import Control.Parallel.Strategies


showNodes :: Network -> Automaton -> String
showNodes net aut = 
    let 
        g = graph aut
     in 
        concatMap (\n -> let p = gLab g n in show n ++ ": " ++ label p ++ " " ++ show (M.toList $ actClks p) ++ "\n") $ G.nodes g 

goalTrim :: Automaton -> Maybe Automaton
goalTrim aut = 
    let g = graph aut
        ns = (\(x, _, _) -> x) $ head $
             dropWhile (not . (\(_, x, y) -> S.null x && S.null y)) $
             iterate (\(ps, xs, ys) -> 
                let pxs = S.map (G.pre g) xs -- set of list of predecessors
                    pys = S.map (map fst . filter (not . S.null . setMapLookup (confForcers aut) . elab . snd) . G.lsuc g) ys -- set of list of successors by (potentially) urgent labels
                    sxs = S.fromList (F.concat pxs) S.\\ ps -- set of new preds
                    sys = S.fromList (F.concat pys) S.\\ ps -- set of new succs st...
                    sfs = S.union sxs sys
                    zs  = S.union xs ys
                 in (S.union ps zs, sxs, sfs) ) (S.empty, goals aut, S.empty)
        autr = aut { graph = G.subgraph (S.toList ns) g, nodeMap = M.filter (\(n, _) -> S.member n ns) (nodeMap aut) }
     in if S.null (goals aut) || not (G.gelem 0 $ graph autr)
           then Nothing
           else Just autr

computeTrace :: Automaton -> Maybe Automaton
computeTrace aut =
    let g   = graph aut
        n   = head $ S.toList $ goals aut
        ns  = n: (predNs . gLab g) n
     in if S.null $ goals aut
           then Nothing
           else Just aut { graph = G.subgraph ns g }

-- Compute the active clocks in each location. Actually map each clock to the
-- maximal constant it will be compared to before the next reset on any path.
updActClocks' :: AGraph -> [(G.Node, M.Map Int Int)] -> AGraph
updActClocks' = foldl' (\g'' (n, cs) -> let pn = gLab g'' n in updateLabel n (pn { actClks = M.unionWith max (actClks pn) cs }) g'') 

updateActClocks :: Automaton -> Automaton
updateActClocks aut = 
    let g = graph aut
        
        -- for each node, a map giving for each clock the max const used in
        -- outgoing edges and in the invariant
        ss   = filter (not . M.null . snd) $ 
               map (\n -> (n, M.fromListWith max $ 
                              S.toList $     
                              S.map guardToxk $                                        -- map guards to pairs (clock, cst) in
                              S.union (maybe S.empty guards $ inv $ gLab g n) $
                              S.unions $                                               -- the big union of 
                              map (\(n', i) -> guards $ fromJust $ timing i) $          -- the guards in
                              filter ((/= Nothing) . timing . snd) $ G.lsuc g n)) $ G.nodes g  -- outgoing edges with a timing

        -- initialise the active clocks
        gi   = updActClocks' g ss

        newg = fst $ head $ dropWhile (not . null . snd) $ 
                   iterate (\(g', ss')  -> 
                          let  
                              -- for each node with its map from clocks to csts
                              -- find the pres and filter the map to keep only clocks not reset
                              -- then I have a big list of (pred node, map)
                              -- I build a map node to union of maps then put it again as a list and
                              -- eliminate the nodes for which the map is empty
                              ss'' = filter (not . M.null . snd) $
                                     M.toList $
                                     M.fromListWith (M.unionWith max) $
                                     concatMap (\(n, m) -> map (\(n', i) -> (n', M.filterWithKey (\k _ -> S.notMember k (maybe S.empty resets $ timing i)) m)) $ 
                                     G.lpre g' n) ss'
                              g''  = updActClocks' g' ss''
                              ss3  = filter (\(n, m) -> actClks (gLab g' n) /= m) ss''
                           in 
                              (g'', ss3)) (gi, ss)
     in
        aut { graph = newg }

-- compute the set of automata that can force a transition in conflict with
-- some transition on the path leading to n.  add also the reqs of the path for
-- efficiency
computeConfReqs :: Automaton -> Int -> S.Set Int
computeConfReqs aut n = 
    let g    = graph aut
        ns   = (predNs . gLab g) n
        cfrs = S.foldl' S.union S.empty $ 
               S.map (setMapLookup $ confForcers aut) $ 
               S.fromList $ 
               concatMap (map (elab . snd) . G.lsuc g) $ 0:ns
     in
        if n == 0 then S.empty
                  else S.difference cfrs $ autoIDs aut

-- Independence of automata
indep :: Automaton -> Automaton -> Bool
indep aut1 = S.null . S.intersection (autoIDs aut1) . autoIDs

-- Effectively update the laReqs and confForcers and resetters, replacing the
-- S.empty placeholder by the real value from lnReqs and labUrgency and
-- clkResets
updatelaReqs :: Network -> Automaton -> Automaton
updatelaReqs net aut = 
    updateActClocks $ aut { 
      laReqs       = M.mapWithKey (\i _ -> M.findWithDefault S.empty i (lnReqs net)) (laReqs aut),
      resetters    = M.mapWithKey (\i _ -> M.findWithDefault S.empty i (clkResets net)) (resetters aut),
      confForcers  = M.filter (not . S.null) $ M.mapWithKey (\i _ -> M.findWithDefault S.empty i (labUrgency net)) (confForcers aut), 
      resetForcers = M.filter (not . S.null) $ M.mapWithKey (\i _ -> M.findWithDefault S.empty i (clockUrgency net)) (resetForcers aut)
    }

-- update goals and automata requirements in automaton number a for objective state s
prepare :: Network -> (Int, Int) -> Automaton
prepare net (a,s) =
    let aut = automata net !! a
        g   = graph aut
        pn  = gLab g s 
        aut' = aut { graph = updateLabel s pn { goal = True } g,
                     goals = S.insert s (goals aut),
                     hasObj = True
                   }
    in autoProduct [updatelaReqs net aut']
        
-- we encode clocks by numbers above the number of automata
-- separate reqs between clocks and autos coming from the path
-- and those coming from conflicts (or resets though it is currently unused)
allReqs :: Automaton -> Int -> G.Node -> (S.Set Int, S.Set Int)
allReqs aut nautos n =
    let
        g       = graph aut
        pn      = gLab g n
        aReqs   = autoReqs pn
        cfReqs  = if M.null $ confForcers aut then S.empty 
                                              else computeConfReqs aut n
        clReqs  = S.map (+nautos) $ clkReqs pn
        rstRqs' = S.foldl' S.union S.empty $ S.map (setMapLookup $ resetForcers aut) $ clkReqs pn
        rstRqs  = if M.null $ resetForcers aut then S.empty
                                               else S.difference rstRqs' $ autoIDs aut
     in 
        (S.union aReqs clReqs, S.union rstRqs cfReqs)


chooseToAdd :: Network -> Automaton -> [(S.Set Int, S.Set Int)] -> ([Int], [Int])
chooseToAdd net aut greqs = 
    let -- we assume that greqs contains no empty set
        nas    = nbautos net
        grs    = map fst greqs -- reqs from the path
        grs'   = map (\(x, y) -> S.union x y) greqs -- reqs also from invariants and external resets
        common = S.toList $ foldl1' S.intersection grs
        mingrs = S.toList $ minimumBy (comparing S.size) grs
        mingrs' = S.toList $ minimumBy (comparing S.size) grs'
        -- if some automaton or clocks is always required for all paths add it
        -- otherwise add the first automaton or clock in the smallest requirement set
        toAdd  = if null common
                     then [ head (if null mingrs then mingrs' else mingrs) ]
                     else common 
        (as, cs) = partition (<nas) toAdd
     in
        (as, map (flip (-) nas) cs)


-- Given the partial product aut, look if adding some more automata is needed (and if so do it) or if the objective cannot be fullfilled (then give Nothing)
concretise :: Network -> Automaton -> Maybe Automaton
concretise net aut = trace ("concretise autos: " ++ show (S.toList $ autoIDs aut) 
                    ++ " -- clocks: " ++ show (S.toList $ aclocks aut) ++ " -- size: " ++ show (G.noNodes $ graph aut) ++ ", " ++ show (length $ G.edges $ graph aut) ++ "\n" -- ++ showAuto net aut
                    )  $ 
    if S.null (goals aut) 
       then Nothing
       else let tra = fromJust $ goalTrim aut -- trim
                nas = nbautos net
                greqs = S.map (allReqs tra nas) $ goals tra
             in 
                if S.member (S.empty, S.empty) greqs
                   then Just tra
                   else
                        let (as, cs) = chooseToAdd net tra $ S.toList greqs
                            auts     = map (updatelaReqs net . (automata net !!)) as
                            noClks   = null cs && S.null (aclocks tra) && all (S.null . aclocks) auts
                            tra'     = if noClks then tra else tra { aclocks = S.union (aclocks tra) (S.fromList cs) }
                            aut'     = if noClks then autoProduct $ tra' : auts 
                                                 else   taProduct $ tra' : auts
                         in 
                            concretise net aut'

syncAutos :: Network -> Automaton -> Automaton -> Maybe Automaton
syncAutos net aut1 aut2 = trace ("sync autos: " ++ show (S.toList $ autoIDs aut1) ++ " x " ++ show (S.toList $ autoIDs aut2) ++ " -- clocks: " ++ show (S.toList $ aclocks aut1) ++ " x " ++ show (S.toList $ aclocks aut2))$
    if indep aut1 aut2 then Just aut2
                       else let aut = if S.null (aclocks aut1) && S.null (aclocks aut2) 
                                         then autoProduct [aut1, aut2]
                                         else taProduct   [aut1, aut2]
                             in -- trace (showAuto net aut)$
                                goalTrim aut >>= concretise net


-- Do the product of overlapping automata
parProduct :: Network -> [Automaton] -> Maybe [Automaton]
parProduct _ []        = Just []
parProduct net (aut:xs) = 
    if all (indep aut) xs
           then (aut:) <$> parProduct net xs
           else mapM (syncAutos net aut) xs >>= parProduct net



-- Possibly give a list of independent partial products of the automata in the Network satisfying the objective and Nothing there exists no such partial product. Concretise first, then product 
preach :: Network -> Maybe [Automaton]
preach net =
    let 
        obj         = objective net
        objAut      = map (concretise net . prepare net) obj -- `using` rpar
     in
        sequence objAut >>= parProduct net

preachFull :: Network -> Maybe [Automaton]
preachFull net = 
    let
        obj   = objective net
        sobj  = S.fromList $ map fst obj
        auts' = map (prepare net) obj ++ map (updatelaReqs net . fst) (filter (\(_, i) -> S.notMember i sobj) $  zip (automata net) [0..])
        auts  = (head auts') { aclocks = S.fromList [0..(length (clocks net) - 1)] } : tail auts'
     in
        sequence [concretise net $ taProduct auts]
