-- This file is part of the reachability analysis tool LaRA
-- (Lazy Reachability Analyzer)
--
-- Loïg Jezequel & Didier Lime
--
-- University of Nantes, École Centrale de Nantes, All rights reserved, 2016

{-# LANGUAGE FlexibleContexts #-}

module Lara.Parser (parseA) where

import qualified Text.Parsec as P
import Data.List (elemIndex, sortBy, foldl1')
import Data.Ord (comparing)
import Data.Maybe

import qualified Data.Foldable as F
import qualified Data.Set as S
import qualified Data.Sequence as Q
import qualified Data.Map.Strict as M
import qualified Data.Sequence as D
import qualified Data.Graph.Inductive.Graph as G

import Control.Monad (when)

import Lara.Data
import Lara.Time

import Debug.Trace

--------------------------------------------------------------------------------
--                                  Parser
--------------------------------------------------------------------------------

data MemParser = MemParser {
                    stateByName :: M.Map String Int,      -- state number by name (current automaton)
                    nameByState :: D.Seq String,          -- state names by number (current automaton)
                    labelByName :: M.Map String Int,      -- label number by name (global)
                    nameByLabel :: D.Seq String,          -- label name by number (global)
                    clockByName :: M.Map String Int,      -- clock number by name (global)
                    nameByClock :: D.Seq String,          -- clock name by number (global)
                    nautos      :: Int,                   -- number of automata read
                    autoByLabel :: M.Map Int (S.Set Int), -- automata using some label by label number
                    labUrg      :: M.Map Int (S.Set Int), -- for each label the set of automata that can force it
                    clockUrg    :: M.Map Int (S.Set Int), -- for each clock the set of automata that can force its reset
                    invByState  :: M.Map Int TimeInfo,    -- invariant for the states that have one (current automaton)
                    cstByClock  :: M.Map Int Int,         -- max cst for each clock (current automaton)
                    autGByClock :: M.Map Int (S.Set Int), -- set of automata using each clock (guard, global)
                    autRByClock :: M.Map Int (S.Set Int)  -- set of automata resetting each clock (global)
                 }

-- Ignore a space block
skipSpaces p = do x<-p; P.spaces; return x

-- Parse a character ignoring trailing spaces
symbol c = P.char c >> P.spaces

-- Parse an identifier: letter then letters or digits
identifier :: P.Parsec String MemParser String
identifier = skipSpaces $ do x<-P.letter; xs<-P.many P.alphaNum; return (x:xs)

-- Parse a number
number :: (Num a, Integral a, Read a) => P.Parsec String MemParser a
number = skipSpaces $ read <$> P.many1 P.digit

-- Parse an invariant
-- returns nothing but updates the memory
pinvariant :: P.Parsec String MemParser ()
pinvariant = skipSpaces $ do
    symbol '['
    x  <- identifier
    
    mem <- P.getState
    let (xi, m') = stateUpdateMem mem x
    P.modifyState (const m')

    ts <- timeInfo

    if (not . S.null . resets) ts || (not. S.null . S.filter guardIsG . guards) ts
        then error "Cannot have resets or 'greater than' constraints in invariants"
        else P.modifyState (\m -> m { invByState = M.insert xi ts (invByState m) } )

    symbol ']'

-- Parse several invariants in an automaton
pinvariants :: P.Parsec String MemParser ()
pinvariants = skipSpaces $ P.skipMany pinvariant 
    

-- Parse a space separated list of clock constraints / resets
timeInfo :: P.Parsec String MemParser TimeInfo
timeInfo = skipSpaces $ do
    x <- identifier
    s <- skipSpaces $ P.choice [ P.string ":=", P.try (P.string "<="), P.try (P.string ">="), P.string "<", P.string ">", P.string "==" ]
    n <- number

    mem <- P.getState
    let (xi, mc) = findOrAdd x $ clockByName mem
    -- if this is a new clock add it
    P.modifyState (\m -> if xi == D.length (nameByClock m) 
                            then m { clockByName = mc, 
                                     nameByClock = nameByClock m D.|> x }
                            else m )

    if s == ":="
       then do P.modifyState (\m -> m { autRByClock = M.insertWith S.union xi (S.singleton $ nautos m) $ autRByClock m } )
               mem <- P.getState
               when (S.size (autRByClock mem M.! xi) > 1) $ error $ "Shared clocks are not allowed: " ++ D.index (nameByClock mem) xi
       else do P.modifyState (\m -> m { autGByClock = M.insertWith S.union xi (S.singleton $ nautos m) $ autGByClock m, 
                                        cstByClock  = M.insertWith max xi n $ cstByClock m } )
               mem <- P.getState
               when (S.size (autGByClock mem M.! xi) > 1) $ error $ "Shared clocks are not allowed: " ++ D.index (nameByClock mem) xi



    case s of
      ":=" -> if n == 0 then return TimeInfo { resets = S.singleton xi, guards = S.empty }
                        else P.unexpected $ "Can only reset clock" ++ x ++ " to zero"
      "<"  -> return TimeInfo { resets = S.empty, guards = S.singleton (GuardL  xi n) }
      "<=" -> return TimeInfo { resets = S.empty, guards = S.singleton (GuardLE xi n) }
      ">"  -> return TimeInfo { resets = S.empty, guards = S.singleton (GuardG  xi n) }
      ">=" -> return TimeInfo { resets = S.empty, guards = S.singleton (GuardGE xi n) }
      "==" -> return TimeInfo { resets = S.empty, guards = S.insert (GuardGE xi n) $ S.singleton (GuardLE  xi n) }
      _    -> error $ "Unknown operator " ++ s




timeInfos :: P.Parsec String MemParser TimeInfo
timeInfos = skipSpaces $ do
    x  <- timeInfo
    xs <- P.option TimeInfo { resets = S.empty, guards = S.empty } timeInfos
    return TimeInfo { resets = S.union (resets x) (resets xs),
                      guards = S.union (guards x) (guards xs) }


-- Parse a space separated list of identifiers 
idSet :: P.Parsec String MemParser (S.Set String)
idSet = skipSpaces $ do s <- identifier; ss <- P.option S.empty idSet; return (S.insert s ss)

-- Parse an automaton
pautomaton :: P.Parsec String MemParser Automaton
pautomaton = do
    skipSpaces (P.string "auto")
    x <- identifier
    symbol '{'

    -- first parse the transitions
    xs <- P.try ptransitions

    -- Then the invariants 
    P.try pinvariants

    symbol '}'

    mem <- P.getState
    let states = nameByState mem
    -- the states that will be put into the graph are ints
    -- we take the first integers
    let nodes = [0..(D.length states - 1)]
    
    -- initially the label of a graph node is: 
    --   - the configuration containing only the node
    --   - not a goal state (to be changed later)
    --   - we do not yet know the automata required to reach it
    --   - the label is the automaton_name.state_name
    --   - the invariant
    let defaultNode n = (n, PNode { config   = ([n], Nothing),
                                    goal     = False, 
                                    autoReqs = S.empty,
                                    clkReqs  = S.empty, 
                                    label    = x ++ "." ++ D.index states n,
                                    predNs   = [],
                                    actClks  = M.empty,
                                    inv      = M.lookup n (invByState mem) })

    -- create the graph
    let g = G.insEdges xs $ G.insNodes (map defaultNode nodes) G.empty

    -- initialize the mapping from configuration to nodes
    let m = M.fromList [([n], (n, Q.empty)) | n <- nodes]

    -- build the default (empty) automata requirements for labels used in this automaton
    let labelSet       = S.fromList $ map (\(_, _, i) -> elab i) xs 
    let ls = M.fromList $ S.toList $ S.map (\i -> (i, S.empty)) labelSet

    -- clocks read (in guards or invariants) in this automaton
    let guardClockSet = S.unions $ map (\(_, _, i) -> mtigClocks $ timing i) xs 
    let invClockSet   = S.unions $ map (mtigClocks . inv . gLab g) nodes
    let cs            = M.fromList $ S.toList $ S.map (\i -> (i, S.empty)) $ S.union guardClockSet invClockSet

    -- filter the transitions with a source invariant
    let ts = filter (\(s, _, _) -> isJust $ M.lookup s $ invByState mem) xs
    -- create a map from urgent labels to the current automaton
    let uls = M.fromList $ map (\(_, _, i) -> (elab i, S.singleton $ nautos mem)) ts
    -- idem for urgent resets
    let urs = M.fromList $ map (\x -> (x, S.singleton $ nautos mem)) $ 
                S.toList $ S.unions $ map (\(_, _, i) -> maybe S.empty resets (timing i)) ts

    -- Add the urgent labels and resets to the memory
    P.modifyState (\m -> m { labUrg   = M.unionWith S.union (labUrg m) uls, 
                             clockUrg = M.unionWith S.union (clockUrg m) urs
                           } )



    -- reset the per-automaton memory: notice how mem is set before this.
    P.modifyState (\m -> m { stateByName = M.empty, 
                             nameByState = D.empty, 
                             cstByClock  = M.empty,
                             invByState  = M.empty,
                             nautos = nautos m + 1 })


    -- Initially:
    --   - this is not a product so contains only itself (nautos mem)
    --   - has no goals (to be updated later) and therefore is not involved in the objective
    return Automaton { graph        = g,
                       nodeMap      = m, 
                       autoIDs      = S.singleton $ nautos mem,
                       goals        = S.empty,
                       name         = x, 
                       laReqs       = ls, 
                       hasObj       = False,
                       aclocks      = S.empty,
                       clockMap     = M.empty,
                       confForcers  = ls,
                       resetForcers = cs,
                       resetters    = cs,
                       kByClock     = cstByClock mem,
                       timed        = False }

-- Parse an indentifier, add it if it does not exists and return its index in the global name list
findOrAdd:: String -> M.Map String Int -> (Int, M.Map String Int)
findOrAdd i ns = 
    let s         = M.size ns
        f _ _ ov  = ov 
        (ni, ns') = M.insertLookupWithKey f i s ns
        in if isNothing ni then (s, ns')
                           else (fromJust ni, ns')

labUpdateMem :: MemParser -> String -> MemParser
labUpdateMem m l = 
    let (xl, ml') = findOrAdd l $ labelByName m
        nlabels   = D.length $ nameByLabel m
     in labUpdateMem' m l nlabels xl ml'

labUpdateMem' :: MemParser -> String -> Int -> Int -> M.Map String Int -> MemParser
labUpdateMem' m l nlabels xl ml' = 
     if xl == nlabels then m { labelByName = ml', 
                               nameByLabel = nameByLabel m D.|> l, 
                               autoByLabel = M.insert xl (S.singleton $ nautos m) (autoByLabel m) } 
                      else m { autoByLabel = M.update (Just . S.insert (nautos m)) xl (autoByLabel m) }

stateUpdateMem :: MemParser -> String -> (Int, MemParser)
stateUpdateMem m l = 
    let (xl, ms') = findOrAdd l $ stateByName m
        nstates = D.length $ nameByState m
     in (xl, if xl == nstates then m { stateByName = ms', 
                                       nameByState = nameByState m D.|> l } 
                              else m)
        


-- Parse a transition
ptransition :: P.Parsec String MemParser (G.LEdge EType)
ptransition = do
    mem <- P.getState

    -- parse a state and add its name to the list if necessary
    i <- identifier
    let (xi, m') = stateUpdateMem mem i
    P.modifyState (const m')

    skipSpaces (P.string "--")
    
    -- parse an edge label and add its name to the list if necessary
    -- also update the set of automata using it
    mem <- P.getState
    l   <- identifier     

    ti  <- P.optionMaybe (skipSpaces $ do symbol '|'; timeInfos)

    P.modifyState (`labUpdateMem` l)

    skipSpaces (P.string "->")
    
    -- Get the possibly updated memory and update nstates
    mem <- P.getState

    -- parse a state and add its name to the list if necessary
    o <- identifier
    let (xo, m'') = stateUpdateMem mem o
    P.modifyState (const m'')

    mem <- P.getState

    -- get the id of the label
    let xl = labelByName mem M.! l

    return (xi, xo, EType { elab = xl, timing = ti })

-- Parse a list of transitions
ptransitions :: P.Parsec String MemParser [G.LEdge EType]
ptransitions = P.many1 ptransition

-- Parse a state description of the form auto.state
pautoloc :: P.Parsec String MemParser (String,String)
pautoloc = do
    a <- identifier
    symbol '.'
    s <- identifier
    return (a, s)

autoByName :: [Automaton] -> String -> Int
autoByName as n =
    let ma = elemIndex n (map name as)
    in fromMaybe (error $ "Automaton " ++ n ++ " unknown") ma
        
autoStateByName :: Automaton -> String -> Int
autoStateByName aut s = 
    let lns   = G.labNodes $ graph aut
        s'    = name aut ++ "." ++ s
        match = filter ((==s').label.snd) lns
    in if null match then error ("State " ++ s ++ " unknown")
                     else fst $ head match
        
-- Parse an objective
pobjective :: [Automaton] -> P.Parsec String MemParser [(Int, Int)]
pobjective as = do
    skipSpaces (P.string "objective")
    os <- pautoloc `P.sepBy1` symbol '&'
    return $ sortBy (comparing fst) $ map (\(x, y) -> let a = autoByName as x in (a, autoStateByName (as !! a) y)) os

-- Parse a list of automata
pautomata :: P.Parsec String MemParser Network
pautomata = do
    as <- P.many1 pautomaton
    -- (0,-1) corresponds to False since it can never be verified
    obj <- P.option (error "An objective must be specified") (pobjective as) 

    mem <- P.getState
    return Network { automata     = as,
                     labels       = F.toList $ nameByLabel mem,
                     clocks       = F.toList $ nameByClock mem,
                     objective    = obj, 
                     lnReqs       = autoByLabel mem,
                     labUrgency   = labUrg mem,
                     clockUrgency = clockUrg mem,
                     clkResets    = autRByClock mem,
                     nbautos      = nautos mem
                   }


-- Main parsing function
parseA :: String -> Either P.ParseError Network
parseA = P.runParser (P.spaces >> pautomata) MemParser { stateByName = M.empty, 
                                                         nameByState = D.empty,
                                                         labelByName = M.empty,
                                                         nameByLabel = D.empty,
                                                         clockByName = M.empty,
                                                         nameByClock = D.empty,
                                                         nautos      = 0,
                                                         autoByLabel = M.empty,
                                                         cstByClock  = M.empty,
                                                         autGByClock = M.empty,
                                                         autRByClock = M.empty,
                                                         labUrg      = M.empty,
                                                         clockUrg    = M.empty,
                                                         invByState  = M.empty
                                                        } "" 


--------------------------------------------------------------------------------
--                              End of Parser
--------------------------------------------------------------------------------


