-- This file is part of the reachability analysis tool LaRA
-- (Lazy Reachability Analyzer)
--
-- Loïg Jezequel & Didier Lime
--
-- University of Nantes, École Centrale de Nantes, All rights reserved, 2016

module Lara.Product 
    (autoProduct,
     taProduct,
     setMapLookup,
     reqsLookup) where

import Data.Ord
import Data.Function
import Data.List
import Data.Maybe
import qualified Data.Graph.Inductive.Graph as G
import qualified Data.Set as S
import qualified Data.Map.Strict as M
import qualified Data.Sequence as Q
import qualified Data.Foldable as F (concat)

import Control.Applicative

import System.Environment
import Debug.Trace

import Lara.Data
import Lara.Time

import Control.Parallel.Strategies

autoCleanUp :: Automaton -> Automaton
autoCleanUp aut =
    let g = graph aut
     in aut {
         -- graph    = G.gmap (\(l, n, pn, r) -> (l, n, ( pn { config = ([n], snd $ config pn) } ), r)) g,
         graph    = G.gmap (\(l, n, pn, r) -> (l, n, pn { config = ([n], Nothing) }, r)) g,
         clockMap = M.empty,
         nodeMap  = M.empty
        } 

-- two small utility functions to shorten the lookup of reqs (conf or la)
setMapLookup :: M.Map Int (S.Set Int) -> Int -> S.Set Int
setMapLookup m i = M.findWithDefault S.empty i m

reqsLookup :: Automaton -> EType -> S.Set Int
reqsLookup aut e = setMapLookup (laReqs aut) $ elab e

timeNext :: Automaton -> G.LNode PNode -> EType -> Maybe TimeInfo -> Maybe DBM
timeNext aut (n, pn) i inv' =
    let 
        g     = graph aut
        cs    = aclocks aut
        cm    = clockMap aut
        etime = tiRemap cm cm $ fromMaybe (TimeInfo S.empty S.empty) $ mtiProject (timing i) cs cs
        invar = tiRemap cm cm $ fromMaybe (TimeInfo S.empty S.empty) $ mtiProject inv' cs cs
        maxK  = M.foldl' max 0 $ M.filterWithKey (\x k -> S.member x $ aclocks aut) $ kByClock aut 
        idbm  = fromJust $ snd $ config pn
     in -- trace ("Next of:\n" ++ show idbm ++ "\nguard: " ++ show etime ++ "\ninvar: " ++ show invar ++ "\nResult:\n" ++ show rdbm ) $ 
        dbmNext idbm etime invar maxK

timeNext' :: Automaton -> G.LNode PNode -> EType -> M.Map Int Int -> (Maybe TimeInfo) -> Maybe DBM
timeNext' aut (n, pn) i act' inv' =
    let 
        g     = graph aut
        act   = actClks pn
        cs    = aclocks aut
        dact  = M.filterWithKey (\k _ -> S.member k cs) act
        dact' = M.filterWithKey (\k _ -> S.member k cs) act'
        kact  = M.keysSet dact
        kact' = M.keysSet dact'
        cmap  = M.fromList $ zip (S.toAscList kact)  [0..]
        cmap' = M.fromList $ zip (S.toAscList kact') [0..]
        etime = tiRemap cmap' cmap  $ fromMaybe (TimeInfo S.empty S.empty) $ mtiProject (timing i) kact' kact
        invar = tiRemap cmap' cmap' $ fromMaybe (TimeInfo S.empty S.empty) $ mtiProject inv'   kact' kact'
        idbm  = fromJust $ snd $ config pn
        --rdbm  = dbmNext' idbm etime invar dact dact'
     in -- trace ("Next of:\n" ++ show idbm ++ "\naclocks " ++ show (aclocks aut) ++ "\nact " ++ show kact ++ "\nact' " ++ show kact' ++"\nguard: " ++ show etime ++ "\ndest invar: " ++ show invar-- ++ "\nResult:\n" ++ show rdbm ) $ 
        dbmNext' idbm etime invar dact dact'


-- product of the p edges (n1,n1',i) of aut1,..., (np,np',i) of autp starting 
-- from node n (which corresponds to (n1,..np) in in the current partial product 
-- aut). Can do nothing if the p edges are not time compatible.
-- return the possibly completed product with possibly the edge added and 
-- wether it leads to a new node or not.
autoProduct1e :: [Automaton] -> [G.LNode PNode] -> EType -> G.LNode PNode -> Automaton -> (Automaton, Maybe (G.LNode PNode), S.Set Int)
autoProduct1e auts tlns i (n, pn) aut =
    let 
        g    = graph aut
        as   = map (S.toList . autoIDs) auts
        sas  = autoIDs aut
        act' = M.unionsWith max $ map (actClks . snd) tlns
        inv' = mtiUnionL $ map (inv . snd) tlns 
        next = timeNext' aut (n, pn) i act' inv'
     in 
        if timed aut && isNothing next
           then -- trace ("-----------------------------------\nfrom " 
                --     ++ show n ++ " " ++ (show $ predNs pn) ++ "\nfire " ++ show i ++ "\n" 
                --     ++ (show $ snd $ config pn) ++ "     -> no succ") $ 
                 (aut, Nothing, S.empty) -- these two edges do not lead to a consistent timed state
           else -- we have a consistent state: does it already exist?
                 let c'      = concatMap (fst . config . snd) tlns
                     mdbm    = if timed aut then next else Nothing
                     newc    = (c', mdbm)
                     passed  = nodeMap aut
                     mn'     = M.lookup c' passed
                     -- auto reqs
                     reqsi = S.difference (reqsLookup aut i) sas
                     reqs' = S.union (autoReqs pn) reqsi                                 
                     -- clock reqs
                     -- clocks in the target invariant except if reset
                     invars'  = S.unions (map (mtigClocks . inv . snd) tlns) S.\\ maybe S.empty resets (timing i)
                     -- plus clocks in the guard and in the source invariant
                     clocksi = S.unions [mtigClocks $ timing i, invars', mtigClocks $ inv pn]
                     clockrq = S.difference clocksi $ aclocks aut
                  in 
                     if isJust mn'
                        then -- it exists: just add an edge and update the list of DBMs
                             let (n', ds)       = fromJust mn'
                                 (ds', b, uids) = if timed aut then dbmAddToList (fromJust next) ds else (Q.empty, False, S.empty)
                                 aut1'          = if G.hasLEdge g (n, n', i) 
                                                     then aut
                                                     else aut { graph = G.insEdge (n, n', i) g }
                                 aut''          = aut1' { nodeMap = M.update (const $ Just (n', ds')) c' (nodeMap aut1') }
                                 pn'            = (gLab g n') { config   = newc, 
                                                                autoReqs = reqs',
                                                                clkReqs  = S.union (clkReqs pn) clockrq,
                                                                predNs   = n : predNs pn
                                                              }
                              in 
                                 if b -- it is timed and we have something new in this DBM
                                    then -- trace ("-----------------------------------\nfrom " 
                                         --    ++ show n ++ " " ++ (show $ predNs pn) ++ "\nfire " ++ show i ++ "\n" 
                                         --    ++ (show $ snd $ config pn) ++ "     -> loop on " ++ show n' ++ " but continue with " ++ show mdbm ) $
                                         (aut'' , Just (n', pn'), uids)
                                         
                                    else -- trace ("-----------------------------------\nfrom " 
                                         --    ++ show n ++ " " ++ (show $ predNs pn) ++ "\nfire " ++ show i ++ "\n" 
                                         --    ++ (show $ snd $ config pn) ++ "     -> loop and stop on " ++ show n' ++ "  with\n" ++ show mdbm) $
                                         (aut'', Nothing, S.empty)
                        else -- it does not exist: add a new node and an edge
                             let n'   = head $ G.newNodes 1 g
                                 -- supposing we start from objectives and add to the right in the product
                                 -- so left always has an objective
                                 goal' = all (\(a, p) -> not (hasObj a) || goal (snd p)) $ zip auts tlns 
                                 -- the label of the new node
                                 pn'   = PNode { config   = newc, 
                                               goal     = goal', 
                                               autoReqs = reqs',
                                               clkReqs  = S.union (clkReqs pn) clockrq,
                                               label    = composeLabels $ map (label . snd) tlns,
                                               predNs   = n : predNs pn,
                                               actClks  = act',
                                               inv      = inv' } -- defined at the very start of the function
                                 -- insert the new edge and node
                                 ds'   = if timed aut then Q.singleton $ fromJust next else Q.empty
                                 aut2' = aut { graph   = G.insEdge (n, n', i) $ G.insNode (n', pn') g,
                                               nodeMap = M.insert c' (n', ds') passed,
                                               goals   = if goal' then S.insert n' (goals aut) 
                                                                  else goals aut }
                              in -- trace ("-----------------------------------\nfrom " 
                                 --    ++ show n ++ " " ++ (show $ predNs pn) ++ "\nfire " ++ show i ++ "\n" 
                                 --    ++ (show $ snd $ config pn) ++ "     -> new node " ++ show n' ++ "\n" ++ (show $ snd $ config pn')) $
                                 (aut2', Just (n', pn'), S.empty)


newtype PWNode = PWNode ([G.LNode PNode], G.LNode PNode) deriving Eq
type PWList = S.Set PWNode

rlength :: PWNode -> Int
rlength (PWNode a) = (S.size . autoReqs . snd . snd) a

pwNotRIncl :: G.LNode PNode -> PWNode -> Bool
pwNotRIncl (n, pn) (PWNode a) = 
    let (n', pn') = snd a
        mb = (liftA2 dbmIncl `on` (snd . config)) pn pn'
     in n /= n' || isNothing mb || (not . fromJust) mb

pwNotInclUid :: S.Set Int -> Int -> PWNode -> Bool
pwNotInclUid uids n (PWNode a) = 
    let (n', pn') = snd a
     in n /= n' || S.notMember ((uid . fromJust . snd . config) pn') uids

instance Ord PWNode where
    a@(PWNode ta) <= b@(PWNode tb) = let xa = rlength a
                                         xb = rlength b
                                     in xa < xb || (xa == xb && ta <= tb)

autoProduct1' :: [Automaton] -> G.LNode PNode -> (Automaton, PWList, Int) -> ([G.LNode PNode], EType) -> (Automaton, PWList, Int)
autoProduct1' auts ln (aut, w, k) (tpns, i) = 
    let (aut', mn, uids) = autoProduct1e auts tpns i ln aut
        (n', pn') = fromJust mn
        (m', md') = config pn'
        ln'       = (n', pn' { config = (m', fmap (\x -> x { uid = k + 1 }) md') })
        neww      = w
        -- neww      = if timed aut then S.filter (pwNotRIncl ln') w else w
        -- neww      = if timed aut && (not . S.null) uids then S.filter (pwNotInclUid uids n') w else w
        w' = if isJust mn then S.insert (PWNode (tpns, ln')) neww else w
        k' = if isJust mn then k + 1 else k
        test = S.map (\(PWNode p) -> uid  $ fromJust $ snd $ config $ snd $ snd p) $ S.filter (not . pwNotRIncl ln') neww 
     in -- trace (if mn /= Nothing then (show $ S.size neww) ++ " <-> " ++ (show $ S.size neww') else "nothing new")$
        -- trace (if mn /= Nothing then show test else "coin") $ 
        (aut', w', k')

autoProduct1 :: [Automaton] -> G.LNode PNode -> (Automaton, PWList, Int) -> [([G.LNode PNode], EType)] -> (Automaton, PWList, Int)
autoProduct1 auts ln = foldl' (autoProduct1' auts ln) 

-- product from two nodes
-- w is the waiting list of nodes to explore in each automata
autoProductN :: [Automaton] -> [G.LNode PNode] ->  G.LNode PNode -> (Automaton, PWList, Int) -> (Automaton, PWList, Int)
autoProductN auts slns ln (aut, w, k) = 
    let 

        -- the list of successors for each automaton
        succs = map (\(a, (n, pn)) -> let g = graph a 
                                          xs = G.lsuc g n 
                                       in map (\(n', i) -> ((n', gLab g n'), i)) xs) $ zip auts slns
        
        -- the base automata involved in each automaton
        reqs  = map autoIDs auts

        -- the list of maps from possible label to possible target nodes
        ls = map (M.fromListWith (++) . map (\(n, i) -> (elab i, [(n, i)]))) succs

        -- the set of all labels from this product node
        labs = S.unions $ map M.keysSet ls
        
        -- complete the map for all labels
        --  - empty if the automaton has this label (but not here)
        --  - ln' if independent
        ls' = map (\(m, ln', r) -> S.foldl' (\m' a -> if M.notMember a m'
                                                         then if S.null $ S.intersection (setMapLookup (laReqs aut) a) r
                                                                 then M.insert a [(ln', EType a Nothing)] m'
                                                                 else M.insert a [] m'
                                                         else m') m labs) $ zip3 ls slns reqs

        -- for each label take the cartesian product of the lists of (target nodes, edge label) for all automata
        es    = concatMap (\a -> mapM (M.! a) ls') $ S.toList labs

        -- compute the union of the edges, and the actual list to add to the waiting list
        edges = map (\xs -> let (ys, zs) = unzip xs in (ys, edgeUnionL zs)) es

        llength (_, x) = S.size $ S.difference (reqsLookup aut x) $ autoIDs aut
     in 
        autoProduct1 auts ln (aut, w, k) $ sortBy (comparing llength) edges
   
autoProduct' :: [Automaton] -> (Automaton, PWList, Int) -> (Automaton, PWList, Int)
autoProduct' auts (aut, w, k) = 
    let PWNode (sns, ln) = S.findMin w
     in if S.null w
           then (aut, S.empty, k)
           else autoProduct' auts $ autoProductN auts sns ln (aut, S.deleteMin w, k)

composeLabels :: [String] -> String
composeLabels = unwords . S.toList . S.unions . map (S.fromList . words)


autoProductInit :: [Automaton] -> Bool -> Automaton
autoProductInit auts timedI = 
    let
        gs       = map graph auts
        pns      = map (`gLab` 0) gs
        as       = S.unions $ map autoIDs auts
        c        = replicate (length auts) 0
        goalI    = all (\(a, p) -> not (hasObj a) || goal p) $ zip auts pns
        clocksI  = S.unions $ map aclocks auts
        actClksI = M.unionsWith max $ map actClks pns
        mdbm     = if timedI then Just dbmi else Nothing
        clkRqI   = S.empty -- S.unions $ map (mtigClocks . inv) pns
        pn       = PNode {
                     config   = (c, mdbm),
                     goal     = goalI,
                     autoReqs = S.empty,
                     clkReqs  = S.difference clkRqI clocksI,
                     label    = composeLabels $ map label pns,
                     predNs   = [],
                     actClks  = actClksI,
                     inv      = mtiUnionL $ map inv pns }
        cmI      = M.fromList $ zip (S.toList clocksI) [0..]
        invarI   = tiRemap cmI cmI $ fromMaybe (TimeInfo S.empty S.empty) $ mtiProject (inv pn) clocksI clocksI
        dbmi     = dbmInit (S.size $ S.intersection clocksI $ M.keysSet actClksI) invarI        
        ds       = if timedI then Q.singleton dbmi else Q.empty
        m        = M.singleton c (0, ds)
        g        = G.insNode (0, pn) G.empty
        goalsI   = if goalI then S.singleton 0 else S.empty
        laReqsI  = M.unions $ map laReqs auts
        cfForceI = M.unions $ map confForcers auts
        clForceI = M.unions $ map resetForcers auts
        resetrsI = M.unions $ map resetters auts
     in 
        Automaton { graph        = g,
                    nodeMap      = m,
                    autoIDs      = as,
                    goals        = goalsI,
                    name         = composeLabels $ map name auts,
                    laReqs       = laReqsI,
                    hasObj       = any hasObj auts,
                    kByClock     = M.unionsWith max $ map kByClock auts, 
                    aclocks      = clocksI,
                    clockMap     = cmI,
                    confForcers  = cfForceI,
                    resetForcers = cfForceI,
                    resetters    = resetrsI,
                    timed        = timedI }


autoProductStart :: [Automaton] -> Automaton -> Automaton
autoProductStart auts aut = 
    let
        lns = map (\a -> (0, gLab (graph a) 0)) auts
     in 
        autoCleanUp $ (\(x, _, _) -> x) $ autoProduct' auts (aut, S.singleton (PWNode (lns, (0, gLab (graph aut) 0))), 0)

-- product of n Automaton's
autoProduct :: [Automaton] -> Automaton
autoProduct auts = autoProductStart auts (autoProductInit auts False)

-- product of two TA's
taProduct :: [Automaton] -> Automaton
taProduct auts = autoProductStart auts (autoProductInit auts True)


