-- This file is part of the reachability analysis tool LaRA
-- (Lazy Reachability Analyzer)
--
-- Loïg Jezequel & Didier Lime
--
-- University of Nantes, École Centrale de Nantes, All rights reserved, 2016

module Lara.Time (
    DBM(..), 
    TimeInfo(..),
    Guard(..),
    mtiUnion,
    mtiUnionL,
    mtiProject,
    mtiProjectNot,
    mtigClocks,
    tiRemap,
    guardClock,
    guardToxk,
    guardBound,
    guardIsG,
    dbmInit, 
    dbmNext,
    dbmNext',
    dbmIncl,
    dbmAddToList
    ) where

import Data.Ord
import Data.Function
import Data.List
import Data.Maybe (fromJust, isNothing)
import qualified Data.Graph.Inductive.Graph as G
import qualified Data.Graph.Inductive.PatriciaTree as GP
import qualified Data.Set as S
import qualified Data.Map.Strict as M
import qualified Data.Sequence as Q
import qualified Data.Vector as V
import qualified Data.Vector.Mutable as VM

import Control.Applicative
import Control.Monad (when, guard)
import Control.Monad.Primitive
import Control.Monad.ST (runST, ST)
import Debug.Trace
import Data.Foldable (forM_)

import qualified Data.Bits as B
import Data.Word (Word64)

--------------------------------------------------------------------------------
--                              DBM Coefficients
--------------------------------------------------------------------------------
data DBMCoeff = Infinity | Finite Int Bool deriving (Eq) -- True is strict

dbmc0 :: DBMCoeff
dbmc0  = Finite 0 False

dbmcInf :: DBMCoeff
dbmcInf = Infinity

instance Ord DBMCoeff where
    _          <= Infinity     = True
    Infinity   <= _            = False
    Finite n s <= Finite n' s' = (n < n') || (n == n' && (s || not s'))

instance Show DBMCoeff where
    show Infinity = "< inf"
    show (Finite n s) = (if s then "<  " else "<= ") ++ show n

dbmcPlus :: DBMCoeff -> DBMCoeff -> DBMCoeff
dbmcPlus Infinity     _              = dbmcInf
dbmcPlus _            Infinity       = dbmcInf
dbmcPlus (Finite n s) (Finite n' s') = Finite (n + n') (s || s') 

dbmcEx :: Int -> DBMCoeff -> DBMCoeff
dbmcEx _ Infinity       = dbmcInf
dbmcEx k c@(Finite n s) 
    | n > k     = dbmcInf
    | n < (-k)  = Finite (-k) True
    | otherwise = c

dbmcEx' :: Int -> Int -> M.Map Int Int -> DBMCoeff -> DBMCoeff
dbmcEx' _ _ _ Infinity        = dbmcInf
dbmcEx' i j m c@(Finite n s)  = 
    let ki = m M.! i
        kj = m M.! j
     in 
        if n > ki then dbmcInf
                  else if n < (-kj) then Finite (-kj) True 
                                    else c

dbmcFromGuard :: Guard -> (Int, Int, DBMCoeff)
dbmcFromGuard (GuardL  x n) = (x + 1, 0, Finite n True)
dbmcFromGuard (GuardLE x n) = (x + 1, 0, Finite n False)
dbmcFromGuard (GuardG  x n) = (0, x + 1, Finite (-n) True)
dbmcFromGuard (GuardGE x n) = (0, x + 1, Finite (-n) False)

dbmcNeg :: DBMCoeff -> DBMCoeff
dbmcNeg Infinity     = error "Cannot negate inf in DBMCoeff"
dbmcNeg (Finite n b) = Finite (-n) (not b)

-- newtype DBMCoeff = DBMC Int deriving Eq
-- 
-- dbmc0 :: DBMCoeff
-- dbmc0  = DBMC 1
-- 
-- dbmcInf :: DBMCoeff
-- dbmcInf = DBMC (maxBound::Int)
-- 
-- dbmcFromInt :: Int -> DBMCoeff
-- dbmcFromInt x = let v = B.shiftL x 1
--                  in if x > 0 && v < 0 || x < 0 && v > 0 
--                        then error $ "overflow in dbmcFromInt: " ++ show x 
--                        else DBMC (B.setBit v 0)
-- 
-- dbmcToInt :: DBMCoeff -> Int
-- dbmcToInt (DBMC x) = B.shiftR x 1
-- 
-- instance Ord DBMCoeff where
--     (DBMC x) <= (DBMC y) = x <= y
-- 
-- instance Show DBMCoeff where
--    show c = if c == dbmcInf 
--                then "< inf" 
--                else (if dbmcIsStrict c then "<  " else "<= ") ++ show (dbmcToInt c)
-- 
-- dbmcMakeStrict :: DBMCoeff -> DBMCoeff
-- dbmcMakeStrict (DBMC x) = DBMC (B.clearBit x 0) 
-- 
-- dbmcIsStrict :: DBMCoeff -> Bool
-- dbmcIsStrict (DBMC x) = not $ B.testBit x 0
-- 
-- dbmcPlus :: DBMCoeff -> DBMCoeff -> DBMCoeff
-- dbmcPlus c d = if c == dbmcInf || d == dbmcInf
--                   then dbmcInf
--                   else let vc = dbmcToInt c
--                            vd = dbmcToInt d 
--                            v  = vc + vd
--                            s  = dbmcIsStrict c || dbmcIsStrict d
--                            r  = dbmcFromInt v
--                         in if (vc >= 0 && vd >= 0 && v < 0) || (vc <= 0 && vd <= 0 && v > 0)
--                               then error $ "DBMCoeff overflow in " ++ show c ++ " + " ++ show d
--                               else if s then dbmcMakeStrict r
--                                         else r
-- 
-- dbmcEx :: Int -> DBMCoeff -> DBMCoeff
-- dbmcEx k c = 
--     if c == dbmcInf
--        then dbmcInf
--        else let v = dbmcToInt c
--              in if v > k 
--                    then dbmcInf
--                    else if v < (-k) then dbmcMakeStrict $ dbmcFromInt (-k)
--                                     else c
-- 
-- dbmcNeg :: DBMCoeff -> DBMCoeff
-- dbmcNeg c = if c == dbmcInf 
--                then error "Cannot negate inf in DBMCoeff"
--                else let d = dbmcFromInt (- (dbmcToInt c))
--                      in if dbmcIsStrict c
--                            then d
--                            else dbmcMakeStrict d
-- 
-- dbmcFromGuard :: Guard -> (Int, Int, DBMCoeff)
-- dbmcFromGuard (GuardL  x n) = (x + 1, 0, dbmcMakeStrict $ dbmcFromInt n)
-- dbmcFromGuard (GuardLE x n) = (x + 1, 0, dbmcFromInt n)
-- dbmcFromGuard (GuardG  x n) = (0, x + 1, dbmcMakeStrict $ dbmcFromInt (-n))
-- dbmcFromGuard (GuardGE x n) = (0, x + 1, dbmcFromInt (-n))

--------------------------------------------------------------------------------
--                                  Guards
--------------------------------------------------------------------------------

data Guard = GuardLE Int Int 
           | GuardGE Int Int 
           | GuardL Int Int 
           | GuardG Int Int 
           deriving (Eq, Ord)

instance Show Guard where
    show (GuardL  n x) = show n ++ "<" ++ show x
    show (GuardLE n x) = show n ++ "<=" ++ show x
    show (GuardG  n x) = show n ++ ">" ++ show x
    show (GuardGE n x) = show n ++ ">=" ++ show x


-- extract the clock in a Guard
guardClock :: Guard -> Int
guardClock (GuardLE x _ ) = x
guardClock (GuardGE x _ ) = x
guardClock (GuardL  x _ ) = x
guardClock (GuardG  x _ ) = x

guardBound :: Guard -> Int
guardBound (GuardLE _ x ) = x
guardBound (GuardGE _ x ) = x
guardBound (GuardL  _ x ) = x
guardBound (GuardG  _ x ) = x


--------------------------------------------------------------------------------
--                                  TimeInfo
--------------------------------------------------------------------------------

data TimeInfo = TimeInfo { resets::S.Set Int, guards::S.Set Guard } deriving (Eq, Ord)

setSerial :: (Ord a, Show a) => S.Set a -> String
setSerial = map (\c -> if c == ' ' then '_' else c) . unwords . S.toList . S.map show

instance Show TimeInfo where
    show (TimeInfo rs gs) = "{" ++ setSerial rs ++ "}_{" ++ setSerial gs ++ "}"

showDBM' :: DBM -> String
showDBM' d = if matrix d == V.empty 
                then "" 
                else show (V.take (size d) (matrix d)) ++ "\n" ++ showDBM' (d { matrix = V.drop (size d) (matrix d) } )

mtigClocks :: Maybe TimeInfo -> S.Set Int
mtigClocks Nothing  = S.empty
mtigClocks (Just t) = S.map guardClock $ guards t


mtiUnion :: Maybe TimeInfo -> Maybe TimeInfo -> Maybe TimeInfo
mtiUnion Nothing Nothing    = Nothing
mtiUnion Nothing mt         = mt
mtiUnion mt      Nothing    = mt
mtiUnion (Just t) (Just t') = 
    let r = S.union (resets t) (resets t')
        g = S.union (guards t) (guards t')
     in Just $ TimeInfo r g

mtiUnionL :: [Maybe TimeInfo] -> Maybe TimeInfo
mtiUnionL [] = error "Cannot take the union of an empty list of mti's"
mtiUnionL xs = foldl1' mtiUnion xs

-- project the time info on clocks not in rs
mtiProjectNot :: Maybe TimeInfo -> S.Set Int -> S.Set Int -> Maybe TimeInfo
mtiProjectNot Nothing  _ _  = Nothing
mtiProjectNot (Just t) rs gs = 
    let r = resets t S.\\ rs
        g = S.filter ((`S.notMember` gs) . guardClock) $ guards t
     in if S.null r && S.null g then Nothing
                                else Just $ TimeInfo r g

-- project the time info on clocks in rs
mtiProject :: Maybe TimeInfo -> S.Set Int -> S.Set Int -> Maybe TimeInfo
mtiProject Nothing  _ _ = Nothing
mtiProject (Just t) rs gs  = 
    let r = S.intersection (resets t) rs
        g = S.filter ((`S.member` gs) . guardClock) $ guards t
     in if S.null r && S.null g then Nothing
                                else Just $ TimeInfo r g


--------------------------------------------------------------------------------
--                                  DBMs
--------------------------------------------------------------------------------

data DBM = DBM { size::Int, matrix::V.Vector DBMCoeff, uid :: Int } deriving (Ord, Eq)

instance Show DBM where
    show = showDBM'

-- all clocks equal and unbounded
dbmInit :: Int -> TimeInfo -> DBM
dbmInit n iv = let 
                   ks = S.map ((\(_, _, c) -> c) . dbmcFromGuard) $ guards iv
                   k = if S.null ks then dbmcInf else S.findMin ks
                   m' = V.replicate (n*n + 2*n +1) dbmc0 V.// [ (i * s, k) | i <- [1.. (s -1)] ]
                   s = n + 1
                in 
                   DBM { size = s, matrix = m' , uid = 0 }

clockRemap :: M.Map Int Int -> Int-> Int
clockRemap m x = M.findWithDefault (error $ "Cannot find clock " ++ show x ++ " in " ++ show m) x m

guardRemap :: M.Map Int Int -> Guard -> Guard
guardRemap m (GuardL  x n) = GuardL   (clockRemap m x) n
guardRemap m (GuardLE x n) = GuardLE  (clockRemap m x) n
guardRemap m (GuardG  x n) = GuardG   (clockRemap m x) n
guardRemap m (GuardGE x n) = GuardGE  (clockRemap m x) n

tiRemap :: M.Map Int Int -> M.Map Int Int -> TimeInfo -> TimeInfo
tiRemap m m' (TimeInfo rs gs) = TimeInfo (S.map (clockRemap m) rs) (S.map (guardRemap m') gs)

guardIsG :: Guard -> Bool
guardIsG (GuardL  _ _) = False
guardIsG (GuardLE _ _) = False
guardIsG (GuardG  _ _) = True
guardIsG (GuardGE _ _) = True

guardToxk :: Guard -> (Int, Int)
guardToxk (GuardL  x n) = (x, n)
guardToxk (GuardLE x n) = (x, n)
guardToxk (GuardG  x n) = (x, n)
guardToxk (GuardGE x n) = (x, n)

_dbmConstrain1 :: PrimMonad m => Int -> VM.MVector (PrimState m) DBMCoeff -> Guard -> m (VM.MVector (PrimState m) DBMCoeff)
_dbmConstrain1 s v g = _dbmConstrain1' s v $ dbmcFromGuard g

_dbmConstrain1' :: PrimMonad m => Int -> VM.MVector (PrimState m) DBMCoeff -> (Int, Int, DBMCoeff) -> m (VM.MVector (PrimState m) DBMCoeff)
_dbmConstrain1' s v (i, j, gij) = 
     do vij <- VM.read v (i * s + j)
        when (gij < vij) $ do
            forM_ [0 .. (s-1)] (\k -> do vik <- VM.read v (i * s + k)
                                         vjk <- VM.read v (j * s + k)
                                         VM.write v (i * s + k) $ min vik $ dbmcPlus gij vjk)
            forM_ [0 .. (s-1)] (\k -> do vki <- VM.read v (k * s + i)
                                         vkj <- VM.read v (k * s + j)
                                         let vki_gij = dbmcPlus vki gij
                                         when (vki_gij < vkj) $ do
                                            VM.write v (k * s + j) vki_gij
                                            forM_ [0 .. (s-1)]
                                                (\l -> do vkl <- VM.read v (k * s + l)
                                                          vjl <- VM.read v (j * s + l)
                                                          VM.write v (k * s + l) $ min vkl $ dbmcPlus vki_gij vjl))

        return v

_dbmConstrain :: PrimMonad m => Int -> VM.MVector (PrimState m) DBMCoeff -> S.Set Guard -> m (VM.MVector (PrimState m) DBMCoeff)
_dbmConstrain s v gs = do forM_ gs $ _dbmConstrain1 s v
                          return v


_dbmFuture :: PrimMonad m => Int -> VM.MVector (PrimState m) DBMCoeff -> m (VM.MVector (PrimState m) DBMCoeff)
_dbmFuture s v = 
    do 
       forM_ [1 .. (s-1)] (\k -> VM.write v (k * s) dbmcInf) 
       return v

_dbmReset1 :: PrimMonad m => Int -> VM.MVector (PrimState m) DBMCoeff -> Int -> m (VM.MVector (PrimState m) DBMCoeff)
_dbmReset1 s v r = 
    do 
       VM.write v ((r + 1) * s) dbmc0
       VM.write v (r + 1)       dbmc0
       forM_ [1 .. (s-1)] (\k -> do v0k <- VM.read v k
                                    vk0 <- VM.read v (k * s)
                                    VM.write v (k * s + r + 1) vk0
                                    VM.write v ((r + 1) * s + k) v0k)  
       return v

_dbmReset :: PrimMonad m => Int -> VM.MVector (PrimState m) DBMCoeff -> S.Set Int -> m (VM.MVector (PrimState m) DBMCoeff)
_dbmReset s v rs = do forM_ rs $ _dbmReset1 s v
                      return v

_dbmExtrapolate :: PrimMonad m => Int -> Int -> VM.MVector (PrimState m) DBMCoeff -> m (VM.MVector (PrimState m) DBMCoeff)
_dbmExtrapolate s k v = do forM_ [0 .. (s * s - 1)] (\i -> do c <- VM.read v i; VM.write v i (dbmcEx k c))
                           return v

_dbmExtrapolate' :: PrimMonad m => Int -> M.Map Int Int -> VM.MVector (PrimState m) DBMCoeff -> m (VM.MVector (PrimState m) DBMCoeff)
_dbmExtrapolate' s m v =
    let 
        ml = M.fromList $ zip (M.keys m) [1..]
        m' = M.insert 0 0 $ M.mapKeys (ml M.!) m
     in do 
            forM_ [0 .. (s - 1)] (\i -> 
                forM_ [0 .. (s - 1)] (\j -> 
                do
                    c <- VM.read v (i * s + j) 
                    VM.write v (i * s + j) (dbmcEx' i j m' c)))

            return v


_dbmEmpty :: (PrimMonad m, Functor m) => Int -> VM.MVector (PrimState m) DBMCoeff -> m Bool
_dbmEmpty s v = any (< dbmc0) <$> mapM (\k -> VM.read v (k * s + k)) [0 .. (s-1)]

dbmNext :: DBM -> TimeInfo -> TimeInfo -> Int -> Maybe DBM
dbmNext d t i k = 
    let s  = size d
        m' = runST $ do m <- V.thaw $ matrix d
                        _dbmConstrain s m $ guards t
                        _dbmReset     s m $ resets t
                        _dbmFuture    s m
                        _dbmConstrain s m $ guards i
                        b <- _dbmEmpty s m
                        if b then return Nothing
                             else do me <- _dbmExtrapolate s k m
                                     fmap Just (V.unsafeFreeze me )
     in fmap (\x -> d { matrix = x}) m'

-- Dynamic DBMs for active clocks
_dbmRedim :: (PrimMonad m, Functor m) => Int -> VM.MVector (PrimState m) DBMCoeff -> M.Map Int Int -> M.Map Int Int -> m (Int, VM.MVector (PrimState m) DBMCoeff)
_dbmRedim s v m m' = 
    let -- keys are in ascending order
        ic = M.fromList $ zip (M.keys m) [1..]
        cs = zip [1..] $ M.keys m'
        -- build a map giving for each new index a maybe old index
        im = M.insert 0 (Just 0) $ M.fromList $ map (\(i, c) -> (i, M.lookup c ic)) cs
        s' = M.size m' + 1
     in 
        do
            v' <- VM.new $ s' * s'
            forM_ [0 .. (s' - 1)] (\i -> do 
                let mi = im M.! i
                if isNothing mi
                then do
                    VM.write v' (i * s') dbmc0
                    VM.write v' i        dbmc0
                else do
                    let i' = fromJust mi
                    v0i <- VM.read v i'
                    vi0 <- VM.read v (i' * s)
                    VM.write v' (i * s') vi0
                    VM.write v' i v0i)

            forM_ [1 .. (s' - 1)] (\i -> do 
                VM.write v' (i * s' + i) dbmc0
                let mi = im M.! i
                if isNothing mi 
                then 
                    forM_ [1 .. (i - 1)] (\j -> do
                        let mj = im M.! j
                        if isNothing mj
                        then do
                           VM.write v' (i * s' + j) dbmc0
                           VM.write v' (j * s' + i) dbmc0
                        else do
                           let j' = fromJust mj
                           v0j <- VM.read v j'
                           vj0 <- VM.read v (j' * s)
                           VM.write v' (i * s' + j) v0j
                           VM.write v' (j * s' + i) vj0)
                else do
                    let i' = fromJust mi
                    forM_ [1 .. (i - 1)] (\j -> do
                        let mj = im M.! j
                        if isNothing mj
                        then do
                           v0i <- VM.read v i'
                           vi0 <- VM.read v (i' * s)
                           VM.write v' (i * s' + j) vi0
                           VM.write v' (j * s' + i) v0i
                        else do
                           let j' = fromJust mj
                           vij <- VM.read v (i' * s + j')
                           vji <- VM.read v (j' * s + i')
                           VM.write v' (i * s' + j) vij
                           VM.write v' (j * s' + i) vji))
            
            return (s', v')


dbmNext' :: DBM -> TimeInfo -> TimeInfo -> M.Map Int Int -> M.Map Int Int -> Maybe DBM
dbmNext' d t i act act' = 
    let s   = size d
        m'' = runST $ do m <- V.thaw $ matrix d
                         _dbmConstrain s m $ guards t
                         (s', m') <- _dbmRedim s m act act'
                         _dbmReset     s' m' $ resets t
                         _dbmFuture    s' m'
                         _dbmConstrain s' m' $ guards i
                         b <- _dbmEmpty s' m'
                         if b then return Nothing
                              -- TODO: extrapolate w/ different constants on each clock
                              -- else do me <- _dbmExtrapolate s' (M.foldl' max 0 act') m'
                              else do me <- _dbmExtrapolate' s' act' m'
                                      fmap Just (V.unsafeFreeze me )
     in 
        fmap (\x -> DBM { size = M.size act' + 1, matrix = x, uid = 0 }) m'' 

-- returns wether d is included or contains d'
dbmIncl' :: Int -> V.Vector DBMCoeff -> V.Vector DBMCoeff -> Int -> Bool
dbmIncl' s m m' i = (i == s * s) || (m V.! i <= m' V.! i && dbmIncl' s m m' (i + 1))

dbmIncl :: DBM -> DBM -> Bool
dbmIncl d d' = 
    let
        m  = matrix d
        m' = matrix d'
        s  = size d
     in 
        dbmIncl' s m m' 0


-- returns wether d is included or contains d'
dbm2Incl :: DBM -> DBM -> (Bool, Bool)
dbm2Incl d d' = 
    let
        m  = matrix d
        m' = matrix d'
     in 
        V.foldr (\(a1, b1) (a2, b2) -> (a1 && a2, b1 && b2)) (True, True) $ V.map  (\(x, x') -> (x <= x', x' <= x)) $ V.zip m m'

-- Add a DBM to a list of DBMs trying to merge.
-- Also returns True if something new is added in the list.
dbmAddToList' :: Q.Seq DBM -> S.Set Int -> DBM -> Q.Seq DBM -> (Q.Seq DBM, Bool, S.Set Int)
dbmAddToList' ds' uids d ds1 =
   if Q.null ds1 
      then (d Q.<| ds', True, uids)
      else 
           let 
               d' Q.:< ds = Q.viewl ds1
               -- (incl, rincl) = dbm2Incl d d'
               incl = dbmIncl d d'
               rincl = dbmIncl d' d
            in
               if incl
                  then (d' Q.<| (ds' Q.>< ds), False, S.empty)     -- inclusion
                  else if rincl  -- reverse inclusion
                           then (d Q.<| (ds' Q.>< ds), True, S.insert (uid d') uids)
                                -- dbmAddToList' ds' (S.insert (uid d') uids) d ds
                           else dbmAddToList' (ds' Q.|> d') uids d ds

--dbmDiff :: Int -> V.Vector DBMCoeff -> V.Vector DBMCoeff -> [(Int, Int, DBMCoeff)]
--dbmDiff s m m' = 
--    let 
--        ns = V.generate (s * s) id
--     in 
--        V.map (\(x, x', i) -> (i `div` s, i `mod` s, x')) $ V.filter (\(x, x', _) -> x > x') $ V.zip3 m m' ns 
--
--refineDiff :: Int -> V.Vector DBMCoeff -> V.Vector (Int, Int, DBMCoeff) -> V.Vector (Int, Int, DBMCoeff) 
--refineDiff s m = V.filter (\(i, j, d) -> let k = i * s + j in m V.! k > d) 
--
--dbmIncl' :: Int -> V.Vector DBMCoeff -> [V.Vector (Int, Int, DBMCoeff)] -> Bool
--dbmIncl' s m [] = False
--dbmIncl' s m (c:cs) =
--    let 
--        cr = refineDiff s m c
--     in 
--        if V.null cr
--           then True
--           else 
--                all (flip (dbmIncl' s) cs) $ dbmSub' s m cr
--
--dbmSub' :: Int -> V.Vector DBMCoeff -> V.Vector (Int, Int, DBMCoeff) -> [V.Vector DBMCoeff]
--dbmSub' s m cs = 
--    V.toList $ V.map (\(i, j, x') -> runST $ do m' <- V.thaw m; mr <- _dbmConstrain1' s m' (j, i, dbmCoeffNeg x'); V.unsafeFreeze mr) cs 
--
--dbmAddToList'' :: DBM -> [DBM] -> ([DBM], Bool)
--dbmAddToList'' d ds =
--    let
--        s   = size d
--        m   = matrix d
--        mds = map matrix ds
--        mcs = map (dbmDiff s m) mds
--        ms  = zip mds mcs 
--     in 
--        if dbmIncl' s m (sortBy (comparing V.length) mcs)
--           then 
--             (ds, False)
--           else 
--             (d:(map (DBM s . fst) $ filter (\(d',cs') -> V.length cs' < (s-1)*s) ms), True)
 

dbmAddToList :: DBM -> Q.Seq DBM -> (Q.Seq DBM, Bool, S.Set Int)
dbmAddToList = dbmAddToList' Q.empty S.empty
--dbmAddToList = dbmAddToList''
